/*  Teutron - A two-headed Tron/Nibbles game
    Copyright (C)2002 Joe Wreschnig <piman@sacredchao.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* This code is really fucking ugly. If someone ever wants to add a high
   score table, make sure to clean it up. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <ctype.h>
#include <curses.h>
#include <getopt.h>

#define UP 1
#define DOWN 2
#define LEFT 3
#define RIGHT 4
#define NONE 5

#define MAX_SPEED 40

#define VERSION "0.6"

typedef struct _Head {
  int dir, speed, score, x, y, counter;
  char is_dead;
} Head;

void move_head(Head *zig) {
  switch (zig->dir) {
  case UP: zig->y--; break;
  case DOWN: zig->y++; break;
  case LEFT: zig->x--; break;
  case RIGHT: zig->x++; break;
  }
}

void check(Head *h, char **board, int *food) {  
  if (h->is_dead) return;
  else if (h->x < 0 || h->y < 0 || h->y >= 20 ||
	   h->x > strlen(board[h->y])) {
    h->is_dead = 1;
    return;
  }

  switch (board[h->y][h->x]) {
  case '-': case '|': case 'o': case 'X': case '#':
    board[h->y][h->x] = 'X';
    h->is_dead = 1;
    h->score -= 500;
    h->dir = NONE;
    break;
  case 's':
    h->speed > MAX_SPEED ? (h->speed = MAX_SPEED) : (h->speed += 3);
    break;
  case 'd':
    h->speed < 3 ? (h->speed = 0) : (h->speed -= 3);
    break;
  case '$':
    h->score += 500;
    break;
  case '*':
    h->score -= 1000;
    break;
  case 'P':
    h->counter = 1000;
    break;
  case '.':
    h->score += 100;
    *food = *food - 1;
    break;
  case '&':{
    int i, j;
    for (i = 0; i < 20; i++)
      for (j = 0; board[i][j] != '\0'; j++)
	if (rand() < (RAND_MAX / 10) && board[i][j] == ' ')
	  board[i][j] = '#';
    break;}
  case '%':{
    int i, j;
    for (i = 0; i < 20; i++)
      for (j = 0; board[i][j] != '\0'; j++)
	  if (board[i][j] == 'o') board[i][j] = ' ';
    break;}
  default:
    if (!h->is_dead) h->score += h->speed / 7;
    break;
  }
  if (board[h->y][h->x] != 'X') board[h->y][h->x] = 'o';
}

void update_screen(char **board, int score, int food) {
  int j;
  char *s = malloc(sizeof(char) * 81);
  for (j = 0; j < 20; j++)
    mvaddstr(j, 0, board[j]);
  refresh();
  snprintf(s, 81, "Teutron %s\t\tScore: %d\t\tFood Left: %d", VERSION, score,
           food);
  mvaddstr(22, 0, s);
}

int load_level(const char *level, char *bonuses, char **board) {
  FILE *f = fopen(level, "r");
  int food = 0;
  char *working;
  int i;
  if (f == NULL) { return -1; }
  fgets(bonuses, 101, f);
  for (i = 0; i < 20 && !feof(f); i++) {
    board[i] = malloc(sizeof(char) * 82);
    fgets(board[i], 82, f);
    working = board[i];
    while ((working = index(working + 1, '.'))) food++;
  }
  for (; i < 20; i++) board[i] = strdup("");
  fclose(f);
  return food;
}


int play_game(const char *level) {
  Head h1, h2;
  int i;
  int c;
  int food = 0;
  char *board[20];
  char *bonuses = malloc(sizeof(char) * 101);
  
  h1.speed = h2.speed = 20;
  h1.score = h2.score = 0;
  h1.counter = h2.counter = MAX_SPEED;
  h1.y = h2.y = 9;
  h1.x = 39; h1.dir = LEFT;
  h2.x = 40; h2.dir = RIGHT;
  h1.is_dead = h2.is_dead = 0;
  
  food = load_level(level, bonuses, board);
  if (food == -1) { return 0; }

  while (!(h1.is_dead && h2.is_dead) && food > 0) {
    while ((c = getch()) != ERR) {
      switch (c) {
      case 'w': if (h1.dir != DOWN) h1.dir = UP; break;
      case 'a': if (h1.dir != RIGHT) h1.dir = LEFT; break;
      case 's': if (h1.dir != UP) h1.dir = DOWN; break;
      case 'd': if (h1.dir != LEFT) h1.dir = RIGHT; break;
	
      case 'i': case KEY_UP: if (h2.dir != DOWN) h2.dir = UP; break;
      case 'j': case KEY_LEFT: if (h2.dir != RIGHT) h2.dir = LEFT; break;
      case 'k': case KEY_DOWN: if (h2.dir != UP) h2.dir = DOWN; break;
      case 'l': case KEY_RIGHT: if (h2.dir != LEFT) h2.dir = RIGHT; break;

      case 'p': nodelay(stdscr, FALSE); getch(); nodelay(stdscr, TRUE); break;
      case 'Q': h1.is_dead = h2.is_dead = 1; break;
      }
    }
    usleep(5000);

    if (rand() % 250 == 1 && strlen(bonuses) != 1) {
      int x = 0, y = 0;
      do {
        y = rand() % 20;
        x = rand() % 80;
      } while (board[y][x] != ' ');
      board[y][x] = bonuses[rand() % (strlen(bonuses) - 1)];
      update_screen(board, h1.score + h2.score, food);
    }
  
    if (h1.counter-- <= h1.speed) {
      h1.counter = MAX_SPEED;
      move_head(&h1); check(&h1, board, &food);
      update_screen(board, h1.score + h2.score, food);
    }
    if (h2.counter-- <= h2.speed) {
      h2.counter = MAX_SPEED;
      move_head(&h2); check(&h2, board, &food);
      update_screen(board, h1.score + h2.score, food);
    }
  }
  for (i = 0; i < 20; i++) free(board[i]);
  free(bonuses);

  if (food != 0)
    if ((h1.score + h2.score) < 0) return (-1000000);
    else return (-1000000 + -1 * (h1.score + h2.score));
  else if (h1.score + h2.score < 0) return 0;
  else return h1.score + h2.score;
}

void print_version() {
  fprintf(stderr, "Teutron %s, Copyright (c)2002 Joe Wreschnig <piman@sacredchao.net>", VERSION);
  fprintf(stderr, "This program is free software; you can redistribute it and/or modify\n");
  fprintf(stderr, "it under the terms of the GNU General Public License as published by\n");
  fprintf(stderr, "the Free Software Foundation; either version 2 of the License, or\n");
  fprintf(stderr, "(at your option) any later version.\n\n");
  fprintf(stderr, "This program is distributed in the hope that it will be useful,\n");
  fprintf(stderr, "but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
  fprintf(stderr, "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
  fprintf(stderr, "GNU General Public License for more details.\n");
  exit(0);
}

void print_help() {
  fprintf(stderr, "Teutron %s - A ridiculously difficult game\n", VERSION);
  fprintf(stderr, "Proper usage: teutron [-h|-v] [-f levelfile]\n");
  fprintf(stderr, "Options:\n");
  fprintf(stderr, " -f [filename]\tLoad this list of levels instead of /etc/teutron.levels\n");
  fprintf(stderr, " -v\t\tDisplay version and copyright information.\n");
  fprintf(stderr, " -h\t\tPrint this help text.\n");
  exit(0);
}

int main(int argc, char **argv) {
  FILE *f;
  char *levels = "/etc/teutron.levels", opt;
  int score = 0, total_score = 0;
  srand(time(NULL));

  while ((opt = getopt(argc, argv, "f:vh")) != -1) {
    switch (opt) {
    case 'f': levels = optarg; break;
    case 'v': print_version(); break;
    case 'h': print_help(); break;
    }
  }

  f = fopen(levels, "r");
  if (f == NULL) {
    endwin();
    fprintf(stderr, "unable to open %s: %s\n", levels, strerror(errno));
    exit(1);
  }
 
  initscr();
  nodelay(stdscr, TRUE);
  keypad(stdscr, TRUE);
  noecho();
  curs_set(0);

  do {
    char *s = malloc(sizeof(char) * PATH_MAX + 1);
    s = fgets(s, PATH_MAX + 1, f);
    if (s != NULL) {
      if (rindex(s, '\n') != NULL) s[rindex(s, '\n') - s] = '\0';
      score = play_game(s);
      if (score > 0) total_score += score;
      else if (score <= -1000000) {
       score += 1000000;
       total_score -= score;
       score = 0;
      } else { total_score -= (score); score = 1; }
    }
    if (s) free(s);
  } while (score != 0 && !feof(f));

  curs_set(1);
  endwin();
  fclose(f);
  printf("Your final score was %d.\n", total_score);
  return 0;
}
